/*----------------------------------------------------------------------
  CRM64Pro GDK - Roberto Prieto
  Copyright (C) 2013-2023 MegaStorm Systems

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

 ------------------------------------------------------------------------ */

#ifndef CRM64PRO_H
#define CRM64PRO_H

// Cross-platform OS and environment support
#if defined(_WIN32) || defined(_WIN64)   
// Windows platforms 
#include <windows.h>
#ifndef _WINDOWS
        #define _WINDOWS
#endif
#ifdef _WIN64
#define _ENV64BIT
#else
        #define _ENV32BIT
#endif
#ifdef CRM64PRO_STATIC        
    // For building/linking static library
    #define HandleDLL		
#elif defined(CRM64PRO_DLL)   
    // For building dynamic library
    #define HandleDLL __declspec(dllexport)
#else                         
    // For linking with dynamic library
    #define HandleDLL __declspec(dllimport)
#endif

#elif defined(__unix__) || defined(__UNIX__)     
    // Android platform
    #ifdef __ANDROID__
        #define _ANDROID
    #else
    // Linux platform
        #define _LINUX
    #endif
    #define HandleDLL 
    #ifdef __LP64__
        #define _ENV64BIT
    #else
        #define _ENV32BIT
    #endif

#elif defined(__MACOSX__) || (__MACH__)
    // MacOS X platform
    #define _MACOSX 
    #define HandleDLL    
    #define _ENV64BIT

#else                   
    #error UNKNOWN PLATFORM - Detected a non-supported OS.
#endif

#if defined(_ENV32BIT) 
    #define CMEM_ALIGNMENT 4
#elif defined(_ENV64BIT)
    #define CMEM_ALIGNMENT 16
#else
#error UNKNOWN ENVIRONMENT - Can not detect if we are running on 32bits or 64bits environment.
#endif

// SDL 2.0.20 check
#include "SDL/SDL.h"  
#if !SDL_VERSION_ATLEAST(2,0,20)
    #error WRONG SDL HEADERS FOUND - SDL version 2.0.20+ must be used.
#endif

// Defines for the fixed Memory modules
#define CMEM_MOD_C64    0x10C64000
#define CMEM_MOD_NET    0x10C64001
#define CMEM_MOD_STL    0x10C64002
#define CMEM_MOD_XML    0x10C64003
#define CMEM_MOD_SDL    0x10C64004
#define CMEM_MOD_MIKMOD 0x10C64005
#define CMEM_MOD_OGG    0x10C64006
#define CMEM_MOD_PNG    0x10C64007
#define CMEM_MOD_ZLIB   0x10C64008
#define CMEM_ENABLED // Enable or disable our Memory Manager and custom STL containers. CRM64Pro and some external libs must be compiled when changing the status.

// STL includes
#include <string>
#include <sstream> 
#include <vector>
#include <deque>
#include <unordered_map>
#include <set>

//! CRM64Pro GDK 
namespace CRM64Pro
{
#ifndef CMEM_ENABLED
	typedef std::basic_string<char, std::char_traits<char>> string;
	typedef std::basic_stringstream<char, std::char_traits<char>> stringstream;
	typedef std::basic_ostringstream<char, std::char_traits<char>> ostringstream;
	typedef std::basic_istringstream<char, std::char_traits<char>> istringstream;
	template <typename T>
	using vector = std::vector<T>;
	template <typename T>
	using deque = std::deque<T>;
	template <typename K, typename T>
	using unordered_map = std::unordered_map<K, T, std::hash<K>, std::equal_to<K>>;
	template <typename K>
	using set = std::set<K, std::less<K>>;
	template <typename K>
	using multiset = std::multiset<K, std::less<K>>;

    // CMemAllocator. All C64 modules inherit from this class.
    class HandleDLL CMemAllocator { };

#else
    // Forward declarations. Some compilers need it.
    namespace CMem
    {
	    HandleDLL void* alloc(size_t, Uint32);
	    HandleDLL void free(void*);
    };
    // Custom STL allocator using our Memory Manager(C++11)
    template <typename T>
    struct stl_allocator
    {
	    typedef T value_type;
	    stl_allocator() = default;
	    template<typename U>
	    stl_allocator(const stl_allocator<U> &) {}
	    stl_allocator(const stl_allocator &) {}
	    stl_allocator & operator=(const stl_allocator &) { return *this; }
	    stl_allocator(stl_allocator &&) = default;
	    stl_allocator & operator=(stl_allocator &&) = default;

	    typedef std::true_type propagate_on_container_copy_assignment;
	    typedef std::true_type propagate_on_container_move_assignment;
	    typedef std::true_type propagate_on_container_swap;

	    bool operator==(const stl_allocator& other) const { return this == &other; }
	    bool operator!=(const stl_allocator& other) const {	return !(*this == other); }

	    T* allocate(std::size_t n) { return static_cast<T*>(CMem::alloc(n * sizeof(T), CMEM_MOD_STL)); }
	    void deallocate(T* ptr, std::size_t n) noexcept	{ CMem::free(ptr); }
    };

    // STL based containers using CMem memory manager    
    typedef std::basic_string<char, std::char_traits<char>, stl_allocator<char> > string;
    typedef std::basic_stringstream<char, std::char_traits<char>, stl_allocator<char> > stringstream;
    typedef std::basic_ostringstream<char, std::char_traits<char>, stl_allocator<char> > ostringstream;
    typedef std::basic_istringstream<char, std::char_traits<char>, stl_allocator<char> > istringstream;
    template <typename T>
    using vector = std::vector<T, stl_allocator<T>>;
    template <typename T>
    using deque = std::deque<T, stl_allocator<T>>;
    template <typename K, typename T>
    using unordered_map = std::unordered_map<K, T, std::hash<K>, std::equal_to<K>, stl_allocator< std::pair<const K, T> >>;
    template <typename K>
    using set = std::set<K, std::less<K>, stl_allocator<K> >;
    template <typename K>
    using multiset = std::multiset<K, std::less<K>, stl_allocator<K> >;

    // CMemAllocator. All C64 modules inherit from this class.
    class HandleDLL CMemAllocator
    {
    public:
        void* operator new(size_t, const std::nothrow_t&);              
        void operator delete(void*, const std::nothrow_t&);
        void operator delete(void*);
        void* operator new[](size_t, const std::nothrow_t&);
        void operator delete[](void*, const std::nothrow_t&);
        void operator delete[](void*);
    };

#endif

// Base class
class HandleDLL Base : public CMemAllocator
{
public:
    virtual Sint32 info(Sint32) = 0;    

protected:
    Base();

private:
    Base(const Base&); // Prohibit the copy constructor
    Base& operator=(const Base&); // Prohibit the operator =         
};

// Manager base class
class HandleDLL ManagerBase : public Base
{
public:
    ManagerBase();
    virtual ~ManagerBase();
    virtual Sint32 create(const string&, Uint32) = 0;
    virtual Sint32 close(Sint32) = 0;
    virtual Sint32 getNum() = 0;
    virtual Sint32 setName(Sint32, const string&) = 0;
};

// Object base class
class HandleDLL ObjectBase : public Base
{
public:
    ObjectBase();
    virtual ~ObjectBase();
    virtual Sint32 getName(string&) = 0;
    virtual Uint32 getID() = 0;
};

// Pre-declare managers
class Main;
class ConfigMgr;
class LogMgr;
class ArchiveMgr;
class XMLMgr;
class Timer;
class Tool;
class ImageMgr;
class CursorMgr;
class TileMgr;
class SpriteMgr;
class FontMgr;
class GFX;
class NetTCP;
class MusicMgr;
class SoundMgr;
class GUIMgr;
class TileEngineMgr;

/** @addtogroup MainM
 *@{ */
/** Event type returned by Main::update() setting one of available codes.*/
#define C64_EVENT          0xEC64 
/** event.user.code returned by Main::update() with C64_EVENT event type. A logic frame occurred.*/
#define C64_EVENT_LOGIC         1
/** event.user.code returned by Main::update() with C64_EVENT event type. A render frame occurred.*/
#define C64_EVENT_RENDER        2
/** event.user.code returned by Main::update() with C64_EVENT event type. A GUI Widget triggered an WS_ACTION event.*/
#define C64_EVENT_WIDGET        4
/** event.user.code returned by Main::update() with C64_EVENT event type. The mouse pointer is over a GUI Widget.*/
#define C64_EVENT_WIDGET_MOUSEOVER  8
/** event.user.code returned by Main::update() with C64_EVENT event type. The GUI Widget lost the focus.*/
#define C64_EVENT_WIDGET_LOSTFOCUS 16
/** General Status: disabled.*/
#define C64_STATUS_DISABLED	0
/** General Status: enabled.*/
#define C64_STATUS_ENABLED	1 
/** General Status: paused.*/
#define C64_STATUS_PAUSED   2
/** General Status: enabled but hidden.*/
#define C64_STATUS_HIDDEN	5
/** General Status: enabled and shown.*/
#define C64_STATUS_SHOWN	9
/** General Status: end.*/
#define C64_STATUS_END     16
/** Error code: Internal error.*/
#define C64_ERR_INTER 0xFFFFFF00
/** Error code: Invalid parameter.*/
#define C64_ERR_PARAM 0xFFFFFE00
/** Error code: Could not open/create a file.*/
#define C64_ERR_FILOP 0xFFFFFD00
/** Error code: Could not write to a file.*/
#define C64_ERR_FILWR 0xFFFFFC00
/** Error code: Could not read from a file.*/
#define C64_ERR_FILRE 0xFFFFFB00
/** Error code: Could not allocate memory.*/
#define C64_ERR_ALLOC 0xFFFFFA00
/** Error code: Object not found.*/
#define C64_ERR_OBJNF 0xFFFFF900

//! Position helpers used by Cursor, Sprite, Font and GUI Widgets
enum ePositionHelpers
{
    PH_END =    0x60000000,     //!< Used with Cursor::setHotSpot() to set as hotspot the cursor_size.<br>Used with Sprite::setAnimHotSpot() to set as hotspot the frame_size.
    PH_CENTER = 0x61000000,     //!< Used with Cursor::setHotSpot() to set as hotspot the cursor_size/2.<br>Used with Sprite::setAnimHotSpot() to set as hotspot the frame_size/2.<br>Widget and Font relative position: center with optional increment.
    PH_CENTERD = 0x71000000,    //!< Widget and Font relative position: center with optional decrement.
    PH_TOP =    0x62000000,     //!< Widget and Font relative position: top with optional increment. Valid for Y axis.
    PH_TOPD =   0x72000000,     //!< Widget and Font relative position: top with optional decrement. Valid for Y axis.
    PH_BOTTOM = 0x63000000,     //!< Widget and Font relative position: bottom with optional increment. Valid for Y axis.
    PH_BOTTOMD = 0x73000000,    //!< Widget and Font relative position: bottom with optional decrement. Valid for Y axis.
    PH_LEFT =   0x65000000,     //!< Widget and Font relative position: left with optional increment. Valid for X axis.
    PH_LEFTD =  0x75000000,     //!< Widget and Font relative position: left with optional decrement. Valid for X axis.
    PH_RIGHT =  0x66000000,     //!< Widget and Font relative position: right with optional increment. Valid for X axis.
    PH_RIGHTD = 0x76000000,     //!< Widget and Font relative position: right with optional decrement. Valid for X axis.
};

//! CMem: CRM64Pro Memory Manager
namespace CMem
{
    /** @addtogroup CMemM
    @{*/
    //! Memory stats gathering level
    enum eMemStatsLevel
    {        
        MSL_NULL = 0,    //!< Disable memory stats gathering. Default value in release code.
        MSL_NORMAL = 2,  //!< Normal memory stats gathering. CRM64Pro modules and user allocations are shown individually. Default value in debug mode.
        MSL_HIGH = 4,    //!< High memory stats gathering. ::MSL_NORMAL level plus deallocations are shown for each module.
        MSL_VERBOSE = 8  //!< Verbose memory stats gathering. ::MSL_HIGH level plus: evolution of allocated memory per module (CMem-Evolution.csv output file) and memory request size histogram per module (CMem-Histogram.csv).
    };
    /** @}*/
    HandleDLL void* alloc(size_t iSize, Uint32 iModule = 0);
    HandleDLL void* alloc_aligned(size_t iSize, size_t iAlign = CMEM_ALIGNMENT, Uint32 iModule = 0);
    HandleDLL void* calloc(size_t iNum, size_t iSize, Uint32 iModule = 0);
    HandleDLL void* realloc(void *pMem, size_t iSize, Uint32 iModule = 0);
    HandleDLL void free(void* pMem);
    HandleDLL void free_aligned(void* pMem);
    HandleDLL void destroy();
    HandleDLL Sint32 info(Sint32 iMode = 1);
    HandleDLL void setLogLevel(eMemStatsLevel eMSL);
    HandleDLL void setVerboseSampleInterval(Sint32 iInterval);
    HandleDLL void setMsgOutput(Sint32 iLM, const char* szFile = nullptr);
    HandleDLL Sint32 setModuleName(Uint32 iModule, const char* szName);
};

//! CRM64Pro GDK Main class
class HandleDLL Main : public Base
{
public:
    static Main& Instance(); // Singleton accessor by object reference
    static void Terminate(); // Explicit singleton destruction
    static Sint32 isInit();    

    // Version and copyrights methods
    Sint32 info(Sint32 iMode = 0);
    void printVersion();
    const string& getVersion();
    const string& getCopyright();
    Sint32 requireVersion(Uint8 iMajor, Uint8 iMinor);

    // Update sytem
    Sint32 update(SDL_Event* evUser = nullptr);
    float getLogicTime();

    // Keyboard input system
    Sint32 getKeyState(SDL_Keycode keycode);
    const char* getKeyName(SDL_Keycode keycode);
    Sint32 getKeyboardFocus();

    // Access to Core components        
    LogMgr& ILogMgr(); // Log Manager        
    ConfigMgr& IConfigMgr(); // Configuration Manager
    ArchiveMgr& IArchiveMgr(); // Archive Manager
    XMLMgr& IXMLMgr(); // XML Manager
    Timer& ITimer(); // Timer interface
    Tool& ITool(); // Tool interface  

    // Access to Audio components
    MusicMgr& IMusicMgr(); // Music Manager
    SoundMgr& ISoundMgr(); // Sound Manager

    // Access to Graphics components
    ImageMgr& IImageMgr(); // Image Manager   
    CursorMgr& ICursorMgr(); // Cursor Manager   
    TileMgr& ITileMgr(); // Tile Manager
    SpriteMgr& ISpriteMgr(); // Sprite Manager
    FontMgr& IFontMgr(); // Font Manager
    GFX& IGFX(); // GFX interface

    // Access to GUI components
    GUIMgr& IGUIMgr(); // GUI Manager

    // Access to TileEngine manager
    TileEngineMgr& ITileEngineMgr();

    // Access to Network components
    NetTCP& INetTCP(); // NetTCP interface

private:
    Main(); // Heap-based and stack-based creation are forbidden.
    ~Main(); // Inheritance is also forbidden.
    static Main* mInstance;
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup LogM
@{*/
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init(), no output at all.*/
#define LM_NULL        0   
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init(), output to a file, if it already exists, its contents are destroyed.*/
#define LM_FILE        2
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init(), output to a file, if it already exists, we write at the file end.*/
#define LM_FILEAPPEND  4
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init, output to standard output stream (stdout).*/
#define LM_STDOUT      8
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init(), output to integrated and default console.*/
#define LM_CONSOLE    16   
/** Log output mode. 'lm' parameter in CRM64Pro::Log::init(), reserved for internal usage, do not use it.*/
#define LM_RESERVED   64  

//! The log level detail
enum eLogLevel
{
    LL_LOW = 0, //!< 'll' parameter in Log::init(), only ::LML_CRITICAL and ::LML_INFO messages are logged out.
    LL_NORMAL = 4, //!< 'll' parameter in Log::init(), previous level messages plus ::LML_NORMAL are logged out.
    LL_HIGH = 16 //!< 'll' parameter in Log::init(), previous level messages plus ::LML_LOW are logged out.
};

//! The importance of a logged message
enum eLogMessageLevel
{
    LML_CRITICAL = 0, //!< 'lml' parameter in Log::msg(), logged on all log levels. Used for critical errors.
    LML_NORMAL = 4,   //!< 'lml' parameter in Log::msg(), logged on ::LL_NORMAL and ::LL_HIGH levels. Used for errors and warnings.
    LML_LOW = 16,     //!< 'lml' parameter in Log::msg(), logged on ::LL_HIGH level. Used for warning and debugging messages.
    LML_INFO = 64,    //!< 'lml' parameter in Log::msg(), logged on all log levels. Timestamp is not supported. Used for logging information messages.
    LML_NULL = 128    //!< 'lml' parameter in Log::msg(), dont log anyting.
};

//! Log Object class
class HandleDLL Log : public ObjectBase
{
    friend class LogMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 init(const string& sAppname, eLogLevel eLL, Sint32 iLM,
                const string& sFilename = "", const string& sAuthor = "",
                const string& sEmail = "", const string& sWeb = "");
    Sint32 pause();
    Sint32 resume();
    Sint32 status();
    Sint32 msg(eLogMessageLevel eLML, char* szFmt,...);
    eLogLevel getLogLevel();
    Sint32 getLogMode();
    Sint32 setLogMode(Sint32 iLM, const string& sFilename = "");
    Sint32 getFilename(string& sName);
    Sint32 setTimeStamp(Sint32 iM);
    Sint32 getTimeStamp();
        
private:
    Log(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Log(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Log Manager class
class HandleDLL LogMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iUnused = 0);
    Sint32 close(Sint32 idLog);
    Sint32 getNum();
    Sint32 setName(Sint32 idLog, const string& sName);
    Log* get(Sint32 idLog = 0);

private:
    LogMgr(); // Heap-based and stack-based creation are forbidden.
    ~LogMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup ConfigM
@{*/
//! Audio drivers
enum eConfigAudioDriver
{
    CAD_DEFAULT = 0, //!< Used in Config::audioSetDriver(), uses the default audio driver (default value).
    CAD_WASAPI = 1, //!< Used in Config::audioSetDriver(), uses WASAPI driver. Only for Win32/Win64.
    CAD_DIRECTSOUND = 2, //!< Used in Config::audioSetDriver(), uses DirectSound driver. Only for Win32/Win64.
    CAD_WINMM = 3, //!< Used in Config::audioSetDriver(), uses Windows Multimedia API. Only for Win32/Win64.
    CAD_ERROR = -1 //!< Returned by Config::audioGetDriver() if the method call failed.        
};

//! Audio frequency
enum eAudioFreq
{
    AF_DEFAULT = 0, //!< Default value.
    AF_NORMAL = 22050, //!< 22050 Hz output.
    AF_HIGH = 44100, //!< 44100 Hz output.
    AF_ULTRA = 48000 //!< 48000 Hz output.
};

//! Audio sample
enum eAudioSample
{
    AS_DEFAULT = 0, //!< Default value.
    AS_8 = AUDIO_S8, //!< 8bits samples output.
    AS_16 = AUDIO_S16SYS, //!< 16bits samples output.
    AS_32 = AUDIO_F32SYS, //!< 32bits floating point samples output. 
};

//! Audio mode
enum eAudioMode
{
    AM_DEFAULT = 0, //!< Default value.
    AM_MONO = 1, //!< Mono output.
    AM_STEREO = 2, //!< Stereo output.
    AM_SURROUND = 4, //!< Surround output.
    AM_SURROUNDPLUS = 6 //!< Surround with center and lfe.
};

//! Screen renderer drivers
enum eConfigRendererDriver
{
    CRD_DEFAULT = 0, //!< Used in Screen::setDriver(), uses the default renderer driver (default value).
    CRD_SOFTWARE = 1, //!< Used in Screen::setDriver(), uses a software renderer driver, it is slow but works on all platforms.
    CRD_OPENGL = 2, //!< Used in Screen::setDriver(), uses OpenGL renderer driver. Supported on Win32/Win64/Linux/MacOS X.
    CRD_OPENGLES2 = 3, //!< Used in Screen::setDriver(), uses OpenGL ES v2.0 renderer driver. Supported on Android/iOS.
    CRD_DIRECT3D = 4, //!< Used in Screen::setDriver(), uses Direct3D v9 renderer driver. Only for Win32/Win64.
    CRD_DIRECT3D11 = 5, //!< Used in Screen::setDriver(), uses Direct3D v11 renderer driver. Only for Win32/Win64 (needs Windows 8+).
    CRD_X11 = 6, //!< Used in Screen::setDriver(), uses X11 renderer driver. Only for Linux.
    CRD_ERROR = -1 //!< Returned by Screen::getDriver() in case of error
};

//! Screen modes
enum eConfigScreenMode
{
    CSM_WINDOW = 0, //!< Used in Screen::setMode(), uses a windowed mode (default value).
    CSM_FULLSCREEN = 1, //!< Used in Screen::setMode(), uses a real fullscreen mode which changes the monitor video mode.
    CSM_FULLSCREENWINDOW = 2 //!< Used in Screen::setMode(), uses a fullscreen window mode which takes the size of the desktop.
};

//! Screen aspect ratio
enum eConfigScreenAspectRatio
{
    CSA_LETTERBOX = 0, //!< Used in Screen::setLogicalSize(), keep the original aspect ratio using(if needed) as filler the required side-bars (default value).
    CSA_OVERSCAN = 1 //!< Used in Screen::setLogicalSize(), expand to fill the entire screen, even if some parts draw off the screen. CRD_DIRECT3D does not support this value.
};

//! Screen Object class
class HandleDLL Screen : public ObjectBase
{
    friend class ConfigMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 show();
    Sint32 hide();
    Sint32 status();
    Sint32 setIcon(SDL_Surface* suIcon);
    Sint32 setDriver(eConfigRendererDriver eCRD);
    eConfigRendererDriver getDriver();
    eConfigRendererDriver getDriverInUse();
    Sint32 setVSync(Uint32 bFlag);
    Uint32 getVSync();
    Sint32 setRenderBatching(Uint32 bFlag);
    Uint32 getRenderBatching();
    Sint32 setMode(eConfigScreenMode eCSM);
    eConfigScreenMode getMode();
    Sint32 setSize(Sint32 iWidth, Sint32 iHeight);
    Sint32 getSize(Sint32* iWidth, Sint32* iHeight);
    Sint32 setLogicalSize(Sint32 iWidth, Sint32 iHeight, eConfigScreenAspectRatio eCSA = CSA_LETTERBOX);
    Sint32 getLogicalSize(Sint32* iWidth, Sint32* iHeight, eConfigScreenAspectRatio* eCSA);
    Sint32 setPosition(Sint32 iX, Sint32 iY);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    Sint32 setClipRect(SDL_Rect* rClip);
    Sint32 getClipRect(SDL_Rect* rClip);
    Sint32 setTitle(const string& sName);
    Sint32 getTitle(string& sName);
    Sint32 setBorder(Uint32 bFlag);
    Uint32 getBorder();
    SDL_Window* getWindow();
    SDL_Renderer* getRenderer();
    SDL_Surface* getSnapshot();
    Sint32 getSnapshot(const string& sFile);
    Sint32 clear(Uint8 iR = 0, Uint8 iG = 0, Uint8 iB = 0, Uint8 iA = 255);
    Sint32 setRenderCallback(Sint32 (*myRenderFunc)(Sint32 iMode, void* pObj), void* pObj = nullptr);
    Sint32 fadeToColor(Uint8 iR, Uint8 iG, Uint8 iB, Sint32 iTime, SDL_Rect* rDst = nullptr);
    Sint32 fadeToImage(Sint32 idImage, Sint32 iTime, SDL_Rect* rDst = nullptr, SDL_Rect* rSrc = nullptr);

private:
    Screen(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Screen(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Config Manager class
class HandleDLL ConfigMgr : public ManagerBase
{
    friend class Main;
public:
    /*! When enabled (1 or greater), it will give back the execution to the operating system scheduler for at less its value in ms after each Render Frame update. By default is disabled (0). */
    Uint8 iMTFriendly;

    // Methods        
    Sint32 info(Sint32 iMode = 0);

    // Audio methods
    Sint32 audioInit(eAudioFreq eAF = AF_DEFAULT, eAudioSample eAS = AS_DEFAULT, eAudioMode eAM = AM_DEFAULT);
    Sint32 audioStatus(eAudioFreq* eAF = nullptr, eAudioSample* eAS = nullptr, eAudioMode* eAM = nullptr);
    Sint32 audioClose();
    Sint32 audioSetDriver(eConfigAudioDriver eCAD);
    eConfigAudioDriver audioGetDriver();
    eConfigAudioDriver audioGetDriverInUse();
    Sint32 audioMusicVolume(Sint32 iVolume);
    Sint32 audioSoundVolume(Sint32 iVolume);
    Sint32 audioMixingChannels(Sint32 iNum);
    Sint32 audioHook(void (*mix_func)(void* udata, Uint8* stream, Sint32 len), void* arg);

    // Screen methods
    Sint32 create(const string& sName, Uint32 iUnused = 0);
    Sint32 close(Sint32 idScreen);
    Sint32 getNum();
    Sint32 setName(Sint32 idScreen, const string& sName);
    void reset(Sint32 iMode = 0);
    Screen* get(Sint32 idScreen = 0);
    Screen* get(SDL_Window* wScreen);

    // Save/load/setup configuration methods    
    Sint32 save(const string& sConfigXML, const string& sCDC = "");
    Sint32 load(const string& sConfigXML, const string& sCDC = "");    
    Sint32 setup(const string& sConfigXML, const string& sConfigCDC = "", const string& sCustomSetupXML = "");
    Sint32 setup(const string& sConfigXML, Sint32 idConfigCDC, const string& sCustomSetupXML = "");

private:
    ConfigMgr(); // Heap-based and stack-based creation are forbidden.
    ~ConfigMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;

    Sint32 renderCallback(Sint32 iMode);
    Sint32 render();
};
/** @}*/

/** @addtogroup ArchiveM
@{*/
//! Archive block storing method
enum eArchiveBlockStore
{
    ABS_RAW = 1, //!< Block is stored in raw mode.
    ABS_COMPRESSED = 2, //!< Block is compressed and stored.
    ABS_CRYPTED = 3 //!< Block is crypted and stored.
};

/** CDC v1.x: size of custom char array attribute 1*/
#define CDCv1_sPARAM1  64
/** CDC v1.x: size of custom char array attribute 2*/
#define CDCv1_sPARAM2  16

//! Archive Object class
class HandleDLL Archive : public ObjectBase
{
    friend class ArchiveMgr;
public:
    /*! Block 
    * CDC common block structure.*/
    struct Block
    {
        /*! Check eArchiveBlockStore enum for further information. Common attribute for all CDC versions.*/
        eArchiveBlockStore eABS;
        /*! Pointer to the raw data. Common attribute for all CDC versions.*/
        void* pData;
        /*! Size in bytes of the data. Common attribute for all CDC versions.*/
        Uint32 iSize;
        /*! Custom attribute Param1 as char array. CDC v1.1 contains the name of the block.*/
        char sParam1[CDCv1_sPARAM1];
        /*! Custom attribute Param2 as char array. CDC v1.1 contains the type of the block.*/
        char sParam2[CDCv1_sPARAM2];
        /*! Custom attribute Param1 as integer. CDC v1.1 contains the stored data size (information purposes only).*/
        Uint32 iParam1;
        /*! Custom attribute Param2 as integer. CDC v1.1 contains the hash of the raw data (information purposes only).*/
        Uint32 iParam2;
        HandleDLL Block();
    };

    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 blockAdd(Block& myBlock);
    Sint32 blockAddFile(const string& sFile, eArchiveBlockStore eABS = ABS_COMPRESSED);
    Sint32 blockDelete(Block& myBlock);
    Sint32 blockUndelete(Uint32 iBlock, Block& myBlock);
    Sint32 blockSearch(Block& myBlock);
    Sint32 blockAt(Uint32 iBlock, Block& myBlock);
    Sint32 blockLoad(Block& myBlock);
    Sint32 blockLoadReleaseData(Block& myBlock);
    Uint32 getUsedBlocks();
    Uint32 getMaxBlocks();
    float getFragmentation();
    Sint32 secure(const unsigned char* myKey, Uint32 iKeySize);
    Sint32 unSecure();

private:
    Archive(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Archive(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Archive Manager class
class HandleDLL ArchiveMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sFile, Uint32 iNumBlocks = 0);
    Sint32 close(Sint32 idCDC);
    Sint32 getNum();
    Sint32 setName(Sint32, const string&); // Unused
    Archive* get(Sint32 idCDC = 0);
    Sint32 defrag(const string& sFile, Uint32 iNumBlocks = 0);
    Sint32 load(const string& sFile, const unsigned char* myKey = nullptr, Uint32 iKeySize = 0);

private:
    ArchiveMgr(); // Heap-based and stack-based creation are forbidden.
    ~ArchiveMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup XMLM
@{*/  
//! XML Object class
class HandleDLL XML : public ObjectBase
{
    friend class XMLMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 nodePointTo(Sint32 iParam, char* szNodes, ...);
    Sint32 nodeFirst();
    Sint32 nodeNext();
    Sint32 nodePrev();
    Sint32 nodeParent();
    Sint32 nodeChild();
    Sint32 nodeStore();
    Sint32 nodeRestore();
    Sint32 nodeRemove();
    Sint32 nodeCreate(const string& sNode);
    Sint32 nodeRename(const string& sNode);
    Sint32 nodeGetName(string& sNode);

    Sint32 setAttribute(const string& sAttr, const string& sValue);
    Sint32 setAttribute(const string& sAttr, const Sint32& iValue);
    Sint32 setAttribute(const string& sAttr, const double& dValue);
    Sint32 getAttribute(const string& sAttr, string& sValue);
    Sint32 getAttribute(const string& sAttr, Sint32& iValue);
    Sint32 getAttribute(const string& sAttr, double& dValue);
    Sint32 removeAttribute(const string& sAttr);

    Sint32 setText(const string& sText);
    Sint32 getText(string& sText);
    Sint32 removeText();

    Sint32 setComment(const string& sComment);
    Sint32 getComment(string& sComment);
    Sint32 removeComment();

    Sint32 saveToFile(const string& sFile);
    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    XML(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~XML(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! XML Manager class
class HandleDLL XMLMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 idXML = 0);
    Sint32 create(const string& sMainNode, Uint32 iUnused = 0);
    Sint32 close(Sint32 idXML);
    Sint32 getNum();
    Sint32 setName(Sint32 idXML, const string& sName);
    XML* get(Sint32 idXML = 0);

    Sint32 loadFromFile(const string& sFile, const string& sName = "");
    Sint32 loadFromBuffer(const unsigned char* szData, const Sint32 iSize, const string& sName);
    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(Sint32 idCDC, const string& sName);
    Sint32 remove(Sint32 idCDC, const string& sName);

private:
    XMLMgr(); // Heap-based and stack-based creation are forbidden.
    ~XMLMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup TimerM
@{*/
//! Timer init state
enum eTimerState
{
    TS_INIT = 0, //!< 'tsOpt' parameter in Timer::init(), initializes or resets all timer settings.
    TS_RESET = 1 //!< 'tsOpt' parameter in Timer::init(), only resets the timer if it was previously initialized keeping RFR and LFR rates.
};

//! Timer class
class HandleDLL Timer : public Base
{
    friend class Main;
public:
    /*! Stats
    * Microbenchmark stats structure.*/
    struct Stats
    {
        /*! Fastest execution time in milliseconds.*/
        double dFastest;
        /*! Slowest execution time in milliseconds.*/
        double dSlowest;
        /*! Range of the execution time in milliseconds.*/
        double dRange;
        /*! Average of the execution time in milliseconds.*/
        double dAverage;
        /*! Median of the execution time in milliseconds.*/
        double dMedian;
        /*! Variance of the execution time in milliseconds.*/
        double dVariance;
        /*! Standard deviation of the execution time in milliseconds.*/
        double dStddev;
        /*! Number of runs.*/
        Uint64 iRuns;
        HandleDLL Stats();
    };

    Sint32 info(Sint32 iMode = 0);
    Sint32 init(eTimerState tsOpt = TS_INIT);
    Sint32 setRate(Uint16 iR, Uint16 iL);
    Sint32 getLFR();
    Sint32 getRFR();
    Uint32 getCurrentRFR();
    Uint32 getCurrentLFR();
    float getAverageRFR();
    float getAverageLFR();
    Uint32 getRenderFrames();
    Uint32 getLogicFrames();
    float getTime();
    Uint32 getTicks();
    Uint32 getTicksNow();
    Uint64 getHiResTime();
    
    Sint32 benchBegin(const string& sName, Sint32 iRuns = 0);
    Sint32 benchBegin(Sint32 idBench);
    Sint32 benchEnd(Sint32 idBench);
    Sint32 benchClose(Sint32 idBench);
    Sint32 benchStats(Sint32 idBench, Stats& stats);    

private:
    Timer(); // Heap-based and stack-based creation are forbidden.
    ~Timer(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update();
    Sint32 incRF();
    Sint32 incLF();
};
/** @}*/

/** @addtogroup ToolM
@{*/
//! Tool MessageBox type
enum eToolMBT
{
    TMBT_EMPTY = 0, //!< MessageBox type empty. 'eTMBT' parameter in Tools:messageBox().
    TMBT_ERROR = 1, //!< MessageBox type error. 'eTMBT' parameter in Tools::messageBox().
    TMBT_WARNING = 2, //!< MessageBox type warning. 'eTMBT' parameter in Tools::messageBox().
    TMBT_INFO = 4 //!< MessageBox type information. 'eTMBT' parameter in Tools::messageBox().
};

/** MessageBox Button OK. 'iButtons' parameter in Tools::messageBox().*/
#define TMBB_OK     1
/** MessageBox Button Cancel. 'iButtons' parameter in Tools::messageBox().*/
#define TMBB_CANCEL 2
/** MessageBox Button Yes. 'iButtons' parameter in Tools::messageBox().*/
#define TMBB_YES    4
/** MessageBox Button No. 'iButtons' parameter in Tools::messageBox().*/
#define TMBB_NO     8

//! Tool class
class HandleDLL Tool : public Base
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Uint32 CRC32(const void* pBuffer, size_t iSize, Uint32 iPreviousCRC32 = 0);
    Uint32 CRC32(const string& sFile);
    Uint32 xxHash(const void* pBuffer, size_t iSize, Uint32 iSeed);
    void* xxHashInit(Uint32 iSeed);
    Sint32 xxHashUpdate(void* pState, const void* pBuffer, size_t iSize);
    Uint32 xxHashDigest(void* pState);
    Sint32 xxHashEnd(void* pState);
    Sint32 XOR(void* pSrc, size_t iSize, Uint8 cKey);
    Sint32 decodeBase64(const string& sInput, string& sOutput, Uint32 iCleanInput = 0);
    Sint32 decompressZlib(const void* pCompressed, Sint32 iCompressedSize, void* pUncompressed, Sint32 iUncompressedSize);
    void randSeedMWC(Uint32 iSeed);
    Uint32 randMWC();
    double randRealMWC();
    void randSeedWELL(Uint32 iSeed, Uint32* pSeedTable = nullptr);
    Uint32 randWELL(Uint32* pSeedState = nullptr, Uint32* pSeedTable = nullptr);
    double randRealWELL(Uint32* pSeedState = nullptr, Uint32* pSeedTable = nullptr);
    float round(const float fNum);    
    double round(const double dNum);
    Sint32 dirExists(const string& sPath);
    Sint32 dirCreate(const string& sPath);
    Sint32 fileExists(const string& sPath);
    Sint32 fileRemove(const string& sPath);
    Sint32 fileGetAbsolutePath(const string& sPath, string& sAbsolutePath);
    Sint32 fileGetExePath(string& sPath);
    Sint32 fileGetName(const string& sPath, string& sFileName);
    Sint32 fileGetExtension(const string& sPath, string& sExtensionName);
    Sint32 fileGetDir(const string& sPath, string& sDirName);    
    Sint32 fileCheckExtension(const string& sPath, const string& sExt);
    Sint32 fileOpenDialog(const string& sTitle, const char* szFilter, string& sSelectedFile);
    Sint32 fileSaveDialog(const string& sTitle, const char* szFilter, string& sSelectedFile);
    Sint32 messageBox(const string& sTitle, const string& sMsg, Uint32 iButtons = TMBB_OK, eToolMBT eTMBT = TMBT_INFO, Sint32 iTTL = 0, Sint32 idCustomPanel = -1, Sint32 idScreen = -1);
    Sint32 szCopy(char* szDst, const char* szSrc, size_t iSize);
    Sint32 getDateStamp(string& sString);
    Sint32 strCompare(const string& sOne, const string& sTwo);
    Sint32 strToLowerCase(string& sString);
    Sint32 strToUpperCase(string& sString);
    Sint32 intToStrDec(Sint32 iNum, string& sString);
    Sint32 intToStrHex(Sint32 iNum, string& sString);
    Sint32 vStrToFile(const vector<string>& vString, const string& sFileName);
    Sint32 fileTovStr(vector<string>& vString, const string& sFileName);

private:
    Tool(); // Heap-based and stack-based creation are forbidden.
    ~Tool(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup Audio
@{*/
//! Audio type
enum eAudioType
{
    AT_nullptr = 0, //!< nullptr type.
    AT_WAV = 2, //!< Wav, Riff, Aiff or Voc audio type.
    AT_MOD = 3, //!< Module audio type.
    AT_OGG = 5 //!< Ogg audio type.
};

//! Audio info 
struct HandleDLL AudioInfo
{
    Sint32 iSize; //!< Audio size in bytes.
    eAudioType eAT; //!< Audio type. Check eAudioType enum.
    Sint32 iFrequency; //!< DSP frequency in samples per second (ogg and wav files).
    Sint32 iFormat; //!< Format: 8, 16 or 32 bits (ogg and wav files).
    Sint32 iChannels; //!< Number of channels (all).
    Sint32 iBitRate; //!< Bitrate (ogg files).
    Sint32 iInstruments; //!< Number of instruments (modules).
    Sint32 iSamples; //!< Number of samples (modules).
    Sint32 iPatterns; //!< Number of patterns (modules).
    Sint32 iDuration; //!< Duration in seconds (all).
    AudioInfo();
};
/** @}*/

/** @addtogroup MusicM
@{*/
//! Music Object class
class HandleDLL Music : public ObjectBase
{
    friend class MusicMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 info(string* sType, AudioInfo* AI);

    Sint32 play(Sint32 iLoop = 0);
    Sint32 fadeIn(Sint32 iTime = 1000, Sint32 iLoop = 0);

    Sint32 saveToFile(const string& sFile);
    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Music(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Music(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Music Manager class
class HandleDLL MusicMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 idMus = 0);
    Sint32 close(Sint32 idMus = 0);
    Sint32 getNum();
    Sint32 setName(Sint32 idMus, const string& sName);
    Music* get(Sint32 idMus = 0);

    Sint32 stop();
    Sint32 pause();
    Sint32 resume();
    Sint32 status();
    Sint32 position(double dPosition);
    Sint32 fadeOut(Sint32 iTime = 1000);

    Sint32 loadFromFile(const string& sFile, const string& sName = "");
    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(const Sint32 idCDC, const string& sName);

private:
    Sint32 create(const string& sMainNode, Uint32 iVersion = 10);
    MusicMgr(); // Heap-based and stack-based creation are forbidden.
    ~MusicMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup SoundM
@{*/
//! Sound Object class
class HandleDLL Sound : public ObjectBase
{
    friend class SoundMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 info(string* sType, AudioInfo* AI);

    Sint32 play(Sint32 iChannel = -1, Sint32 iLoop = 0);
    Sint32 fadeIn(Sint32 iTime = 1000, Sint32 iChannel = -1, Sint32 iLoop = 0);

    Sint32 saveToFile(const string& sFile);
    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Sound(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Sound(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Sound Manager class
class HandleDLL SoundMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 idSnd = 0);
    Sint32 close(Sint32 idSnd = 0);
    Sint32 getNum();
    Sint32 setName(Sint32 idSnd, const string& sName);
    Sound* get(Sint32 idSnd = 0);

    Sint32 stop(Sint32 iChannel);
    Sint32 pause(Sint32 iChannel);
    Sint32 resume(Sint32 iChannel);
    Sint32 status(Sint32 iChannel);
    Sint32 fadeOut(Sint32 iChannel, Sint32 iTime = 1000);

    Sint32 panning(Sint32 iChannel = -1, Uint8 iLeft = 255, Uint8 iRight = 255);
    Sint32 distance(Sint32 iChannel = -1, Uint8 iDistance = 0);
    Sint32 position(Sint32 iChannel = -1, Sint16 iAngle = 0, Uint8 iDistance = 0);

    Sint32 loadFromFile(const string& sFile, const string& sName = "");
    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(const Sint32 idCDC, const string& sName);

private:
    Sint32 create(const string& sMainNode, Uint32 iVersion = 10);
    SoundMgr(); // Heap-based and stack-based creation are forbidden.
    ~SoundMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup ImageM
@{*/
//! Image filters
enum eImageFilter
{
    IF_NOISE = 1, //!< 'eFilter' parameter in Image::applyFilter. Apply a noise effect with a specific iValue (from 0 to 100).
    IF_BLUR = 2, //!< 'eFilter' parameter in Image::applyFilter. Apply a blur effect with a specific iValue (from 0 to 100).
    IF_NEGATIVE = 3, //!< 'eFilter' parameter in Image::applyFilter. Apply a negative effect.
    IF_GREY = 4, //!< 'eFilter' parameter in Image::applyFilter. Apply a grey effect.
    IF_BW = 5, //!< 'eFilter' parameter in Image::applyFilter. Apply a black/white effect.
    IF_WIGGLE = 6, //!< 'eFilter' parameter in Image::applyFilter. Apply a wiggle effect that can be animated using variable iValue.
    IF_PIXELATE = 7 //!< 'eFilter' parameter in Image::applyFilter. Apply a pixelate effect that can be animated using variable iValue.
};

//! Image Object class
class HandleDLL Image : public ObjectBase
{
    friend class ImageMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Sint32 assignSurface(SDL_Surface* pSurf, Sint32 iOwnership = 0);
    Sint32 setOwner(Sint32 iOwner);
    Sint32 getOwner();

    Sint32 optimize(Sint32 idScreen = 0);
    Sint32 rebuild();
    Sint32 render(Sint32 idRes = 0, SDL_Rect* rSrc = nullptr, SDL_Rect* rDst = nullptr);
    Sint32 renderEx(const double dAngle, const SDL_Point* poCenter = nullptr, const SDL_RendererFlip rf = SDL_FLIP_NONE,
                    Sint32 idRes = 0, SDL_Rect* rSrc = nullptr, SDL_Rect* rDst = nullptr);
    Sint32 applyFilter(eImageFilter eFilter, Sint32 iValue = 0);

    Sint32 setBlendMode(SDL_BlendMode bmValue);
    SDL_BlendMode getBlendMode();
    Sint32 setAlphaMod(Sint32 iAlpha);
    Sint32 getAlphaMod();
    Sint32 setColorMod(Uint8 iR, Uint8 iG, Uint8 iB);
    Sint32 getColorMod(Uint8& iR, Uint8& iG, Uint8& iB);
    Sint32 setColorKey(Sint16 iR = -1, Sint16 iG = -1, Sint16 iB = -1);
    Sint32 getColorKey(Sint16& iR, Sint16& iG, Sint16& iB);

    SDL_Surface* getSurface();
    Sint32 findTex(Sint32& iIndex, SDL_Texture*& pTex, Sint32& idScreen);

    Sint32 saveToBMP(const string& sFile = "");
    Sint32 saveToPNG(const string& sFile = "");
    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Image(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Image(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Image Manager class
class HandleDLL ImageMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 idImg = 0);
    Sint32 create(const string& sName, Uint32 iVersion = 10);
    Sint32 close(Sint32 idImg);
    Sint32 getNum();
    Sint32 setName(Sint32 idImg, const string& sName);
    Sint32 closeEx(Sint32 idImg, Uint32 iOwner);
    Image* get(Sint32 idImg);

    Sint32 duplicate(Sint32 idImg, const string& sName);
    Sint32 child(Sint32 idImg);

    Sint32 isBMP(const string& sFile);
    Sint32 isPNG(const string& sFile);

    Sint32 loadFromFile(const string& sFile, const string& sName = "");
    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 load(void* pBuffer, Sint32 iSize, const string& sName);
    Sint32 remove(Sint32 idCDC, const string& sName);

private:
    ImageMgr(); // Heap-based and stack-based creation are forbidden.
    ~ImageMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup CursorM
@{*/
//! Cursor Object class
class HandleDLL Cursor : public ObjectBase
{
    friend class CursorMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 assignImage(Sint32 idImage, Sint32 iOwnership = 0);
    Sint32 getImage();
    Sint32 build();
    Sint32 setHotSpot(Sint32 iHotX, Sint32 iHotY);
    Sint32 getHotSpot(Sint32* iHotX, Sint32* iHotY);

    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Cursor(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Cursor(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Cursor Manager class
class HandleDLL CursorMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iVersion = 10);
    Sint32 close(Sint32 idCur);
    Sint32 getNum();
    Sint32 setName(Sint32 idCur, const string& sName);
    Cursor* get(Sint32 idCur);

    Sint32 select(Sint32 idCur);
    Sint32 getSelect();
    Sint32 show();
    Sint32 hide();
    Sint32 setPosition(Sint32 iX, Sint32 iY, Sint32 idScreen = 0);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    Sint32 getButtons();
    Sint32 getButtonsPrev();
    Sint32 getFocus();

    Sint32 child(Sint32 idCur);

    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(Sint32 idCDC, const string& sName);

private:
    CursorMgr(); // Heap-based and stack-based creation are forbidden.
    ~CursorMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update();
};
/** @}*/

/** @addtogroup TileM
@{*/
//! Tile Object class
class HandleDLL Tile : public ObjectBase
{
    friend class TileMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 assignImage(Sint32 idImage, Sint32 iOwnership = 0);
    Sint32 getImage();

    Sint32 setOffset(Sint32 iTSMode, Sint32 iX, Sint32 iY);
    Sint32 getOffset(Sint32* iX, Sint32* iY);
    Sint32 setPosition(Sint32 iX, Sint32 iY);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    Sint32 getNumTiles(Sint32* iX = nullptr, Sint32* iY = nullptr);
    Sint32 render(Sint32 iTS = 1, Sint32 idRes = 0);
    Sint32 renderEx(Sint32 iTS = 1, const double dAngle = 0.0, const SDL_Point* poCenter = nullptr, const SDL_RendererFlip rf = SDL_FLIP_NONE, Sint32 idRes = 0);

    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Tile(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Tile(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Tile Manager class
class HandleDLL TileMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iVersion = 10);
    Sint32 close(Sint32 idTile);
    Sint32 getNum();
    Sint32 setName(Sint32 idTile, const string& sName);
    Tile* get(Sint32 idTile);

    Sint32 child(Sint32 idTile);

    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(const Sint32 idCDC, const string& sName);

private:
    TileMgr(); // Heap-based and stack-based creation are forbidden.
    ~TileMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup SpriteM
@{*/
/** Sprite animation type: loop.*/
#define SPR_TYPE_LOOP      1
/** Sprite animation type: ping-pong.*/
#define SPR_TYPE_PINGPONG  2
/** Sprite animation type: one-shot.*/
#define SPR_TYPE_ONESHOT   4
/** Sprite animation high-state: normal     (10000000-00000000)*/
#define SPR_STATE_NORMAL      0x8000 
/** Sprite animation high-state: up         (01000000-00000000)*/
#define SPR_STATE_UP          0x4000 
/** Sprite animation high-state: down       (00100000-00000000)*/
#define SPR_STATE_DOWN        0x2000 
/** Sprite animation high-state: left       (00010000-00000000)*/
#define SPR_STATE_LEFT        0x1000 
/** Sprite animation high-state: right      (00001000-00000000)*/
#define SPR_STATE_RIGHT       0x0800 
/** Sprite animation high-state: up-left    (00000100-00000000)*/
#define SPR_STATE_UPLEFT      0x0400 
/** Sprite animation high-state: up-right   (00000010-00000000)*/
#define SPR_STATE_UPRIGHT     0x0200 
/** Sprite animation high-state: down-left  (00000001-00000000)*/
#define SPR_STATE_DOWNLEFT    0x0100 
/** Sprite animation high-state: down-right (11000000-00000000)*/
#define SPR_STATE_DOWNRIGHT   0xC000 
/** Each state is a 16bit value, the high byte set the high-state which is defined by the GDK.<br>Use this macro for getting the sprite animation high-state. See above SPR_STATE_xx.*/
#define SPR_GET_HIGHSTATE(x) ((x) & 0xFF00) 
/** Each state is a 16bit value, the low byte set the low-state(up to 256) which can be defined by the user.<br>Use this macro for getting the sprite animation low-state.*/
#define SPR_GET_LOWSTATE(x)  ((x) & 0x00FF) 

//! Sprite Object class
class HandleDLL Sprite : public ObjectBase
{
    friend class SpriteMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 assignImage(Sint32 idImage, Sint32 iOwnership = 0);
    Sint32 getImage();

    Sint32 setOffset(Sint32 iX, Sint32 iY);
    Sint32 getOffset(Sint32* iX, Sint32* iY);
    Sint32 pause();
    Sint32 resume();
    Sint32 status();
    Sint32 setSpeed(float fSpeed);
    Sint32 getAnimNumber();
    Sint32 getAnimCurrent();
    Sint32 selectAnim(Sint32 iAnim);
    Sint32 setAnimType(Sint32 iAnim, Sint32 iType);
    Sint32 getAnimType(Sint32 iAnim, Sint32* iType);
    Sint32 setAnimState(Sint32 iAnim, Sint32 iState);
    Sint32 getAnimState(Sint32 iAnim, Sint32* iState);
    Sint32 orderAnimState(Sint32 iState = -1);
    Sint32 setAnimHotSpot(Sint32 iAnim, Sint32 iHotX, Sint32 iHotY);
    Sint32 getAnimHotSpot(Sint32 iAnim, Sint32* iHotX, Sint32* iHotY);    
    Sint32 setAnimRange(Sint32 iAnim, Sint32 iFrameStart, Sint32 iFrameEnd);
    Sint32 getAnimRange(Sint32 iAnim, Sint32* iFrameStart, Sint32* iFrameEnd);
    Sint32 getFrameNumber(Sint32 iAnim);
    Sint32 getFrameCurrent();
    Sint32 selectFrame(Sint32 iAnim, Sint32 iFrame);
    Sint32 setFrameTime(Sint32 iAnim, Sint32 iFrame, Uint16 iTime);
    Sint32 getFrameTime(Sint32 iAnim, Sint32 iFrame, Uint16* iTime);

    Sint32 setPosition(Sint32 iX, Sint32 iY, Sint32 iInterpolation = 0);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    Sint32 render(Sint32 idRes = 0);
    Sint32 renderEx(const double dAngle = 0.0, const SDL_Point* poCenter = nullptr, const SDL_RendererFlip rf = SDL_FLIP_NONE, Sint32 idRes = 0);

    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Sprite(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Sprite(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Sprite Manager class
class HandleDLL SpriteMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iVersion = 10);
    Sint32 close(Sint32 idSprite);
    Sint32 getNum();
    Sint32 setName(Sint32 idSprite, const string& sName);
    Sprite* get(Sint32 idSprite);

    Sint32 child(Sint32 idSprite);

    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(const Sint32 idCDC, const string& sName);

private:
    SpriteMgr(); // Heap-based and stack-based creation are forbidden.
    ~SpriteMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup FontM
@{*/
//! Font Object class
class HandleDLL Font : public ObjectBase
{
    friend class FontMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 assignImage(Sint32 idImage, Sint32 iOwnership = 0);
    Sint32 getImage();

    Sint32 setCursor(Uint8 cCursor);
    Uint8 getCursor();
    Sint32 getCursorPositionByPixel(const string& sText, Sint32 iPixelPos);
    Sint32 setKerning(Sint32 iKer);
    Sint32 getKerning();
    Sint32 getWidth(const string& sText, Sint32 iCurPos = -1);
    Sint32 getHeight();

    Sint32 setPosition(Sint32 iX, Sint32 iY);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    Sint32 render(const string& sText, Sint32 iCurShow = 0, Sint32 iCurPos = -1, Sint32 idRes = 0);
    Sint32 renderEx(const string& sText, Sint32 iCurShow = 0, Sint32 iCurPos = -1,
                    const double dAngle = 0.0, const SDL_Point* poCenter = nullptr, const SDL_RendererFlip rf = SDL_FLIP_NONE, Sint32 idRes = 0);

    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);

private:
    Font(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Font(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! Font Manager class
class HandleDLL FontMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iVersion = 10);
    Sint32 close(Sint32 idFont);
    Sint32 getNum();
    Sint32 setName(Sint32 idFont, const string& sName);
    Font* get(Sint32 idFont);

    Sint32 child(Sint32 idFont);

    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(const Sint32 idCDC, const string& sName);
    Sint32 remove(const Sint32 idCDC, const string& sName);

private:
    FontMgr(); // Heap-based and stack-based creation are forbidden.
    ~FontMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup GFXM
@{*/
//! GFX class
class HandleDLL GFX : public Base
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);

    Sint32 setSurfacePixel(Sint32 iX, Sint32 iY, Uint32 iColor, SDL_Surface* pSurf);
    Uint32 getSurfacePixel(Sint32 iX, Sint32 iY, SDL_Surface* pSurf);

    Sint32 setPixel(Sint32 iX, Sint32 iY, Uint32 iColor, Sint32 idRes = 0);
    Sint32 setPixel(Sint32 iX, Sint32 iY, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Uint32 getPixel(Sint32 iX, Sint32 iY, Sint32 idRes = 0);

    Sint32 lineH(Sint32 iX1, Sint32 iX2, Sint32 iY, Uint32 iColor, Sint32 idRes = 0);
    Sint32 lineH(Sint32 iX1, Sint32 iX2, Sint32 iY, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 lineV(Sint32 iX, Sint32 iY1, Sint32 iY2, Uint32 iColor, Sint32 idRes = 0);
    Sint32 lineV(Sint32 iX, Sint32 iY1, Sint32 iY2, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 line(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint32 iColor, Sint32 idRes = 0);
    Sint32 line(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 rect(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint32 iColor, Sint32 idRes = 0);
    Sint32 rect(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 rectFilled(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint32 iColor, Sint32 idRes = 0);
    Sint32 rectFilled(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 rectRound(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Sint32 iRad, Uint32 iColor, Sint32 idRes = 0);
    Sint32 rectRound(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Sint32 iRad, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 rectFilledRound(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Sint32 iRad, Uint32 iColor, Sint32 idRes = 0);
    Sint32 rectFilledRound(Sint32 iX1, Sint32 iY1, Sint32 iX2, Sint32 iY2, Sint32 iRad, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 arc(Sint32 iX, Sint32 iY, Sint32 iRad, Sint32 iStart, Sint32 iEnd, Uint32 iColor, Sint32 idRes = 0);
    Sint32 arc(Sint32 iX, Sint32 iY, Sint32 iRad, Sint32 iStart, Sint32 iEnd, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 circle(Sint32 iX, Sint32 iY, Sint32 iRad, Uint32 iColor, Sint32 idRes = 0);
    Sint32 circle(Sint32 iX, Sint32 iY, Sint32 iRad, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 circleFilled(Sint32 iX, Sint32 iY, Sint32 iRad, Uint32 iColor, Sint32 idRes = 0);
    Sint32 circleFilled(Sint32 iX, Sint32 iY, Sint32 iRad, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 ellipse(Sint32 iX, Sint32 iY, Sint32 iRadX, Sint32 iRadY, Uint32 iColor, Sint32 idRes = 0);
    Sint32 ellipse(Sint32 iX, Sint32 iY, Sint32 iRadX, Sint32 iRadY, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 ellipseFilled(Sint32 iX, Sint32 iY, Sint32 iRadX, Sint32 iRadY, Uint32 iColor, Sint32 idRes = 0);
    Sint32 ellipseFilled(Sint32 iX, Sint32 iY, Sint32 iRadX, Sint32 iRadY, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 polygon(Sint32* iVX, Sint32* iVY, Sint32 iNV, Uint32 iColor, Sint32 idRes = 0);
    Sint32 polygon(Sint32* iVX, Sint32* iVY, Sint32 iNV, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);
    Sint32 polygonFilled(Sint32* iVX, Sint32* iVY, Sint32 iNV, Uint32 iColor, Sint32 idRes = 0);
    Sint32 polygonFilled(Sint32* iVX, Sint32* iVY, Sint32 iNV, Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, Sint32 idRes = 0);

    Sint32 RGBtoHSV(Uint8 iR, Uint8 iG, Uint8 iB, float* fH, float* fS, float* fV);
    Sint32 HSVtoRGB(float fH, float fS, float fV, Uint8* iR, Uint8* iG, Uint8* iB);
    Sint32 RGBtoYUV(Uint8 iR, Uint8 iG, Uint8 iB, Uint8* iY, Uint8* iU, Uint8* iV);
    Sint32 YUVtoRGB(Uint8 iY, Uint8 iU, Uint8 iV, Uint8* iR, Uint8* iG, Uint8* iB);
    Sint32 RGBtoCMYK(Uint8 iR, Uint8 iG, Uint8 iB, Uint8* iC, Uint8* iM, Uint8* iY, Uint8* iK);
    Sint32 CMYKtoRGB(Uint8 iC, Uint8 iM, Uint8 iY, Uint8 iK, Uint8* iR, Uint8* iG, Uint8* iB);
    Sint32 RGBtoYCC(Uint8 iR, Uint8 iG, Uint8 iB, Uint8* iY, Uint8* Cb, Uint8* Cr);
    Sint32 YCCtoRGB(Uint8 iY, Uint8 Cb, Uint8 Cr, Uint8* iR, Uint8* iG, Uint8* iB);

private:
    GFX(); // Heap-based and stack-based creation are forbidden.
    ~GFX(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup GUIWidgetM
@{*/
//! Widget state
enum eWidgetState
{
    WS_DEFAULT = -1, //!< Flag used for "setters" methods. Not a real state.
    WS_NORMAL = 0,  //!< Normal widget state.
    WS_HOVERED = 1, //!< Mouse pointer is over the widget.
    WS_PRESSED = 2, //!< Pressed widget state.
    WS_ACTION = 3,  //!< Action widget state.
    WS_DEACTIVATED = 4 //!< Deactivated widget state.
};

//! Widget type
enum eWidgetType
{
    WT_BASE = 0, //!< Base widget.
    WT_LABEL = 1, //!< Simple label widget.
    WT_IMAGE = 2, //!< Simple image widget.
    WT_BUTTON = 4, //!< Standard button widget.
    WT_CHECKBOX = 8, //!< CheckBox widget.        
    WT_PROGRESSBAR = 16, //!< ProgressBar widget.		
    WT_HSLIDER = 32, //!< Horizontal slider widget.
    WT_VSLIDER = 64, //!< Vertical slider widget.
    WT_TEXTBOX = 128, //!< TextBox widget.
    WT_LISTBOX = 256 //!< ListBox widget.		
};

/** Widget feature: fading effect at showing and hiding time.*/
#define WIDGET_FEATURE_FADE      0x00010000
/** Widget feature: allow to move the widget (drag&drop).*/
#define WIDGET_FEATURE_DRAGDROP  0x00020000
/** Widget feature: enclose the widget inside its parent (default).*/
#define WIDGET_FEATURE_ENCLOSE   0x00040000
/** Widget feature: the widget is detached from its parent.*/
#define WIDGET_FEATURE_DETACH    0x00080000
/** Widget feature: background image/sprite auto-fit to widget size.*/
#define WIDGET_FEATURE_BGFIT     0x00100000
/** Widget feature: background rendering is disabled.*/
#define WIDGET_FEATURE_BGDISABLE 0x00200000
/** Widget feature: allow the widget to process events and generate C64_EVENT_WIDGET event.*/
#define WIDGET_FEATURE_EVENTS    0x00400000
/** Widget feature: allow the widget to generate C64_EVENT_WIDGET_MOUSEOVER events.*/
#define WIDGET_FEATURE_MOUSEOVER 0x00800000
/** Widget feature: allow the widget to generate C64_EVENT_WIDGET_LOSTFOCUS events.*/
#define WIDGET_FEATURE_LOSTFOCUS 0x01000000
/** Widget feature: TextBox supports multiline text.*/
#define WIDGET_FEATURE_MULTILINE 0x02000000
/** Widget feature: TextBox on read-only mode.*/
#define WIDGET_FEATURE_READONLY  0x04000000
/** Widget feature: ListBox auto-sorting of items.*/
#define WIDGET_FEATURE_AUTOSORT  0x08000000

// Pre-declare Panel
class Panel;

//! Widget Object
class HandleDLL Widget : public CMemAllocator
{
    friend class Panel;
    friend class Console;
    friend class DebugWindow;
public:
    virtual Sint32 info(Sint32 iMode = 0);
    Sint32 disable();
    Sint32 show();
    Sint32 hide();
    Sint32 status();
    Sint32 features(Sint32 iFeatures);

    virtual Sint32 setSize(Sint32 iWidth, Sint32 iHeight);
    Sint32 getSize(Sint32* iWidth, Sint32* iHeight);
    Sint32 setPosition(Sint32 iX = PH_CENTER, Sint32 iY = PH_CENTER);
    Sint32 getPosition(Sint32* iX, Sint32* iY);
    virtual Sint32 setMargin(Sint32 iTop, Sint32 iRight, Sint32 iBottom, Sint32 iLeft);
    Sint32 getMargin(Sint32* iTop, Sint32* iRight, Sint32* iBottom, Sint32* iLeft);
    Sint32 setAlphaMod(Uint8 iAlpha);
    Uint8 getAlphaMod();
    virtual Sint32 setText(const string& sText, Sint32 iParam1 = PH_CENTER, Sint32 iParam2 = PH_CENTER);
    virtual Sint32 getText(string& sText, Sint32 iParam1 = -1);
    virtual Sint32 setKey(SDL_Keycode keycode);
    SDL_Keycode getKey();
    virtual Sint32 setValue(Sint32 iValue);
    virtual Sint32 getValue();
    virtual Sint32 getValueGroup(Sint32 iGroup);
    virtual Sint32 setRange(Sint32 iRange);
    virtual Sint32 getRange();
    Sint32 setState(eWidgetState eWS = WS_NORMAL);
    virtual Sint32 setScroll(Sint32 iUnitsX, Sint32 iUnitsY);
    virtual Sint32 getScroll(Sint32* iUnitsX, Sint32* iUnitsY);
    virtual Sint32 getVisibleItems(Sint32* iUnitsX, Sint32* iUnitsY);
    virtual Sint32 setFont(Sint32 idFont, eWidgetState eWS = WS_DEFAULT);
    Sint32 getFont(eWidgetState eWS = WS_NORMAL);
    virtual Sint32 setGroup(Sint32 iGroup);
    virtual Sint32 getGroup();    
    Sint32 setBorderWidth(Sint32 iWidth, eWidgetState eWS = WS_DEFAULT);
    Sint32 getBorderWidth(eWidgetState eWS = WS_NORMAL);
    Sint32 setBorderCorner(Sint32 iRad, eWidgetState eWS = WS_DEFAULT);
    Sint32 getBorderCorner(eWidgetState eWS = WS_NORMAL);
    Sint32 setBorderColor(Uint8 iR, Uint8 iG, Uint8 iB, Uint8 iA, eWidgetState eWS = WS_DEFAULT);
    Sint32 getBorderColor(Uint8* iR, Uint8* iG, Uint8* iB, Uint8* iA, eWidgetState eWS = WS_NORMAL);
    Sint32 setBgColor(Uint8 iR, Uint8 iG, Uint8 iB, eWidgetState eWS = WS_DEFAULT);
    Sint32 getBgColor(Uint8* iR, Uint8* iG, Uint8* iB, eWidgetState eWS = WS_NORMAL);
    Sint32 setBgImage(const string& sFileCDC, const string& sName, eWidgetState eWS = WS_NORMAL);
    Sint32 setBgImage(Sint32 idCDC, const string& sName, eWidgetState eWS = WS_NORMAL);
    Sint32 setBgImage(Sint32 idImage, eWidgetState eWS = WS_NORMAL);
    Sint32 getBgImage(eWidgetState eWS = WS_NORMAL);
    Sint32 setBgSprite(const string& sFileCDC, const string& sName, eWidgetState eWS = WS_DEFAULT, Sint32 iAnim = 0);
    Sint32 setBgSprite(Sint32 idCDC, const string& sName, eWidgetState eWS = WS_DEFAULT, Sint32 iAnim = 0);
    Sint32 setBgSprite(Sint32 idSprite, eWidgetState eWS = WS_DEFAULT, Sint32 iAnim = 0);
    Sint32 getBgSprite(eWidgetState eWS = WS_NORMAL, Sint32* iAnim = nullptr);
    Sint32 setFunction(Sint32 (*stateFunc)(void*, void*), void* iParam1 = nullptr, void* iParam2 = nullptr, eWidgetState eWS = WS_ACTION);

protected:
    Widget(Uint32, const string&, Panel*); // Heap-based and stack-based creation are forbidden.
    virtual ~Widget(); // Inheritance is also forbidden.
    Widget(const Widget&); // Prohibit the copy constructor
    Widget& operator=(const Widget&); // Prohibit the operator = 
    class PIMPL;
    PIMPL* mPData;
    virtual Sint32 update(SDL_Event*, Sint32);
    virtual Sint32 render(Sint32, Sint32);
    Sint32 save(Sint32 idCDC);
};
/** @}*/

/** @addtogroup GUIM
@{*/
/** GUI Object: Panel.*/
#define GUI_PANEL        0x42000000
/** GUI Object: Console.*/
#define GUI_CONSOLE      0x44000000
/** GUI Object: DebugWindow.*/
#define GUI_DEBUGWINDOW  0x48000000

//! Panel Type
enum ePanelType
{
    PT_RETRIEVE = -1, //!< Used for getting current panel type.
    PT_MODELESS = 0, //!< Modeless panel. When it gets the focus, it does not block the input for others panels.
    PT_MODAL = 1, //!< Modal panel. When it gets the focus, it blocks the input for others panels on the same screen.
    PT_EPHEMERAL = 2 //!< Ephemeral and modeless panel. It lasts the given time duration.
};

//! Panel Object class
class HandleDLL Panel : public ObjectBase
{
    friend class GUIMgr;
    friend class WidgetCheckBox;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Widget& baseWidget();
    Sint32 attachTo(Sint32 idScreen);
    Sint32 getScreen();

    Sint32 createWidget(const string& sName, eWidgetType eWT, Sint32 iWID);
    Widget* getWidget(Sint32 iWID);
    Sint32 closeWidget(Sint32 iWID);
    Sint32 type(ePanelType ePT = PT_RETRIEVE, Sint32 iTime = 0);

    Sint32 save(const string& sFileCDC);
    Sint32 save(Sint32 idCDC);      

private:
    Panel(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Panel(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update(SDL_Event*, Sint32);
    Sint32 render();
    Sint32 wgCBGroup(Sint32, Sint32);
    Sint32 loadWidget(Archive*, const string&);
};

//! Console Object class
class HandleDLL Console : public ObjectBase
{
    friend class GUIMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Widget& baseWidget();
    Sint32 attachTo(Sint32 idScreen);

    Sint32 print(const char* format,...);
    Sint32 vPrint(const char* format, va_list args);
    Sint32 addCmdHandler(const string& sCommandName, const string& sCommandHelp, Sint32 (*cmdFunc)(vector<string>*, Console* pCon));
    Sint32 executeCmd(const string&);

private:
    Console(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~Console(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update(SDL_Event*, Sint32);
    Sint32 render();
};

//! DebugWindow Object class
class HandleDLL DebugWindow : public ObjectBase
{
    friend class GUIMgr;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 getName(string& sName);
    Uint32 getID();
    Widget& baseWidget();
    Sint32 attachTo(Sint32 idScreen);

    Sint32 addWatch(const string& sName, Sint32* iIntegerVar, double* dDoubleVar = nullptr);
    Sint32 removeWatch(const string& sName);
    Sint32 setRefreshInterval(Sint32);
    Sint32 getRefreshInterval();

private:
    DebugWindow(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~DebugWindow(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update(SDL_Event*, Sint32);
    Sint32 render();
};

//! GUI Manager class
class HandleDLL GUIMgr : public ManagerBase
{
    friend class Main;
    friend class Console;
public:
    Sint32 info(Sint32 idGUI = 0);
    Sint32 create(const string& sName, Uint32 iType = GUI_PANEL);
    Sint32 close(Sint32 idGUI);
    Sint32 getNum();
    Sint32 setName(Sint32 idGUI, const string& sName);
    Panel* getPanel(Sint32 idGUI = 0);
    Sint32 getPanelFocus();
    Sint32 setPanelFocus(Sint32 idGUI);
    Console* getConsole(Sint32 idGUI = GUI_CONSOLE);
    DebugWindow* getDebugWindow(Sint32 idGUI = 0);

    Sint32 load(const string& sFileCDC, const string& sName);
    Sint32 load(Sint32 idCDC, const string& sName);

    Sint32 getFontWhiteAA();
    Sint32 getFontBlack();
    Sint32 getIconInfo();
    Sint32 getIconWarning();
    Sint32 getIconError();

private:
    GUIMgr(); // Heap-based and stack-based creation are forbidden.
    ~GUIMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
    Sint32 update(SDL_Event*);
    Sint32 render();
    Sint32 cmdCon(vector<string>*, Console*);
    Sint32 cmdDw(vector<string>*, Console*);
};
/** @}*/

/** @addtogroup TE
@{*/
/** Disable all layer's flags. Default value.*/
#define TE_LAYERFLAG_DISABLE		 0 
/** Enable the loop on X axis.*/
#define TE_LAYERFLAG_LOOPX			 1 
/** Enable the loop on Y axis.*/
#define TE_LAYERFLAG_LOOPY			 2 
/** Enable autoscroll flag. Check setLayerAutoScroll() for further details.*/
#define TE_LAYERFLAG_AUTOSCROLL		 4 
/** Enable the updating flag. Check update() for further details.*/
#define TE_LAYERFLAG_UPDATE			 8 
/** Enable the updating cells flag. Check update() for further details.*/
#define TE_LAYERFLAG_UPDATECELLS    16 
/** Enable the rendering flag. Check render() for further details.*/
#define TE_LAYERFLAG_RENDER			32 
/** Enable the smooth rendering flag. Check render() for further details.*/
#define TE_LAYERFLAG_SMOOTHSCROLL	64 
/** Show the base mapdata grid (for debugging purposes). Check setLayerGridColor() for further details.*/
#define TE_LAYERFLAG_SHOWGRID	   128 
/** Pause the layer (current positions can not be modified). Note that update() and render() methods should be called to keep the internal state.*/
#define TE_LAYERFLAG_PAUSE		   256 

/**  Used with properties members to avoid to change the current value.*/
#define TE_KEEP_VALUE           0x00FFFFFE 
/**  Set the layer position at the beginning.*/
#define TE_LAYERPOSITION_BEGIN  0x00FFFFFD 
/**  Set the layer position at the middle.*/
#define TE_LAYERPOSITION_MIDDLE 0x00FFFFFC 
/**  Set the layer position at the end.*/
#define TE_LAYERPOSITION_END    0x00FFFFFB 
/**  Try to set the layer position at the center.*/
#define TE_LAYERPOSITION_CENTER 0x0A000000 

/** Maximum registered mapdata types. [0] is always cMSTE_MapBase.*/
#define TE_REGISTEREDMAPDATA_MAX		 3 
/** Default source for loaded tilesets. It sets that at saving level time, the tileset will be saved in the same location as the level.*/
#define TE_TILESET_DEFAULT_SOURCE		"TE-TS-DEFAULT-SOURCE" 
/** Default name for a layer which is allocated but empty.*/
#define TE_LAYER_EMPTY					"TE-LAYER-EMTPY" 

//! TE Resource control
enum eTileEngineResCtrl
{
    TERC_USER = 0, //!< Used for layer maps and tilesets. The user must remove these resources after usage. Can be modified for layers but not for tilesets.
    TERC_AUTO = 1, //!< Used for layer maps and tilesets. The engine will automatically remove these resources. Can be modified for layers but not for tilesets.
    TERC_INTERNAL = 2 //!< Used for layer maps and tilesets. Importing or loading levels will set on this flag. The engine automatically manages the resources. Can not be modified.
};
/** @}*/

/** @addtogroup TEMapM
@{*/
//! MapData abstract class definition.
class HandleDLL TE_MapData : public CMemAllocator
{
public:
    // Methods for dynamic 2D arrays management
    template < typename T >  T **create2DArray(Sint32 height, Sint32 width);
    template < typename T >  Sint32  delete2DArray(T** Array);

    // General stuff methods
    TE_MapData();
    virtual ~TE_MapData();

    // Pure virtual Map Access methods (public)
    virtual Sint32 checkMap() = 0;
    virtual Sint32 info(Sint8, eLogMessageLevel LML) = 0;
    virtual Sint32 cellUpdate(Sint32, Sint32, void*, void*) = 0;
    virtual Sint32 cellRender(Sint32, Sint32, Sint32, Sint32, Sint32, Sint32, Sint32, Sint32) = 0;
    virtual Sint32 saveMapdataNode(Sint32, Sint32) = 0;
    virtual Sint32 loadMapdataNode(Sint32, Sint32) = 0;

    // Properties methods
    Sint32 getMapType(string& sMT);
    Sint32 getCellWidth();
    Sint32 setCellWidth(Sint32 size);
    Sint32 getCellHeight();
    Sint32 setCellHeight(Sint32 size);
    Sint32 getMapWidth();
    Sint32 getMapHeight();

protected:
    class PIMPL;
    PIMPL* mPData;
};

//! Template: dynamic 2D array of a given data type.
/*! \param height Selected height of your 2D array
\param width Selected width of your 2D array
\return A Pointer to the dynamic 2D array*/
template < typename T > T** TE_MapData::create2DArray(Sint32 height, Sint32 width)
{
    T **ppi;
    T *pool;
    T *curPtr;

    #ifdef CMEM_ENABLED
        // 1.Allocate memory for array of elements of column
        ppi = (T**)CMem::alloc(sizeof(T*) * height, CMEM_MOD_C64);
        // 2.Allocate memory for array of elements of each row
        pool = (T*)CMem::alloc(sizeof(T) * width * height);
    #else
        // 1.Allocate memory for array of elements of column
        ppi = new(std::nothrow) T*[height];
        // 2.Allocate memory for array of elements of each row
        pool = new(std::nothrow) T[width * height];
    #endif

    // 3.Now point the pointers in the right place
    curPtr = pool;
    for(int i = 0; i < height; i++)
    {
        *(ppi + i) = curPtr;
        curPtr += width;
    }
    // 4.Return
    return ppi;
}

//! Template: delete a dynamic 2D array.
/*!
\param Array pointer to 2D array.
\return 0 on success.*/
template < typename T > Sint32 TE_MapData::delete2DArray(T** Array)
{
    #ifdef CMEM_ENABLED
        CMem::free(*Array);
        CMem::free(Array);
    #else
        delete[]* Array;
        delete[] Array;
    #endif
    return 0;
}

//! MapBase class. Minimum mapdata to be used with the TileEngine. Inherit from it your own custom MapData.
class HandleDLL TE_MapBase : public TE_MapData
{
public:

    // General stuff methods
    TE_MapBase();
    TE_MapBase(Sint32 height, Sint32 width);
    ~TE_MapBase();
    static TE_MapBase *createInstance(Sint32 height, Sint32 width); // Create an instance of this class.

    // Definition of Map Access methods for this kind of map (public)
    Sint32 checkMap();
    Sint32 info(Sint8, eLogMessageLevel eLML);
    Sint32 cellUpdate(Sint32, Sint32, void*, void*);
    Sint32 cellRender(Sint32, Sint32, Sint32, Sint32, Sint32, Sint32, Sint32, Sint32);
    Sint32 saveMapdataNode(Sint32 idXML, Sint32 iOutput);
    Sint32 loadMapdataNode(Sint32 idXML, Sint32 iInput);

    // MapBase methods
    Sint32 createMapBase(Sint32 height, Sint32 width, Sint16 value = -1);
    Sint32 deleteMapBase();
    Sint32 setMapBaseValue(Sint32 y, Sint32 x, Sint16 value);
    Sint32 cloneMapBase(TE_MapBase *);

    // Dynamic 2D array for the tilemap
    Uint32 **iTilemap;

protected:
    // Protected methods
    Sint32 A2DtoCSV(Uint32 **Array2D, Sint32 w, Sint32 h, char* &szOutput);
    Sint32 CSVtoA2D(Uint32 **Array2D, Sint32 w, Sint32 h, const string& sInput);
    Sint32 stringtoA2D(Uint32 **Array2D, Sint32 w, Sint32 h, const char* szInput);
};
/** @}*/

/** @addtogroup TECoreM
@{*/
//! TileEngine Object class
class HandleDLL TileEngine : public ObjectBase
{
    friend class TileEngineMgr;
public:
    typedef Sint32(*TE_OnLayerEnd)(Sint32 cellX, Sint32 cellY, Sint32 startX, Sint32 endX, Sint32 incX, Sint32 startY, Sint32 endY, Sint32 incY);

    Sint32 info(Sint32 iMode = -1);
    Sint32 getName(string& sName);
    Uint32 getID();

    Sint32 allocateLayers(Sint32 iN = -1);
    Sint32 freeLayer(Sint32 layer);
    Sint32 freeLevel();
    Sint32 logVerbosity(Sint32);

    Sint32 setViewport(SDL_Rect *rect);
    SDL_Rect* getViewport();
    Sint32 getModificationDate(string& sDate);
    Sint32 getCreationDate(string& sDate);
    Sint32 registerMapDataHandler(TE_MapBase* (*createInstance)(Sint32 y, Sint32 x));
    Sint32 addCustomMapData(Sint32 layer, TE_MapBase* map);
    Sint32 deleteCustomMapData(Sint32 layer);
    
    Sint32 loadTileset(Sint32 tileID, Sint32 iID = -1);
    Sint32 loadTileset(const string& sCDCfile, const string& sTileName, Sint32 iID = -1);
    Sint32 loadTileset(const string& sIMGfile, const string& sTileName, Sint32 iTileWidth, Sint32 iTileHeight, Sint32 iID = -1);
    Sint32 unloadTileset(Sint32 iTN);
    Sint32 getTileset(Sint32 iTN);
    Sint32 getTilesetSource(Sint32 iTN, string& sName);
    Sint32 setTilesetSource(Sint32 iTN, const string& sSource);

    Sint32 setLayerFlags(Sint32 layer, Sint32 iFlags, char iState = 0);
    Sint32 getLayerFlags(Sint32 layer);
    Sint32 setLayerResControl(Sint32 layer, eTileEngineResCtrl rescontrol);
    Sint32 getLayerResControl(Sint32 layer);
    Sint32 setLayerAlphaMod(Sint32 layer, Sint32 alpha);
    Sint32 getLayerAlphaMod(Sint32 layer);
    Sint32 setLayerParallaxRatio(Sint32 layer, float pratio);
    float getLayerParallaxRatio(Sint32 layer);
    Sint32 setLayerPosition(Sint32 layer, float xpos = TE_KEEP_VALUE, float ypos = TE_KEEP_VALUE);
    Sint32 getLayerPosition(Sint32 layer, float* xpos, float* ypos);
    Sint32 setLayerAutoScroll(Sint32 layer, float xspeed = TE_KEEP_VALUE, float yspeed = TE_KEEP_VALUE);
    Sint32 getLayerAutoScroll(Sint32 layer, float* xspeed, float* yspeed);
    Sint32 setLayerMapData(Sint32 layer, TE_MapBase* map);
    TE_MapBase* getLayerMapData(Sint32 layer);
    Sint32 setLayerOnUpdateEnd(Sint32 layer, TE_OnLayerEnd myfunc);
    TE_OnLayerEnd* getLayerOnUpdateEnd(Sint32 layer);
    Sint32 setLayerOnRenderEnd(Sint32 layer, TE_OnLayerEnd myfunc);
    TE_OnLayerEnd* getLayerOnRenderEnd(Sint32 layer);
    Sint32 getLayerName(Sint32 layer, string& sName);
    Sint32 setLayerName(Sint32 layer, const string& sName);
    Sint32 setLayerGridColor(Sint32 layer, Uint32 iColor);   

    Sint32 screen2LayerCell(Sint32 layer, Sint32 sx, Sint32 sy, Sint32* cx, Sint32* cy);
    Sint32 screen2LayerAbsolute(Sint32 layer, Sint32 sx, Sint32 sy, Sint32* px, Sint32* py);
    Sint32 moveLayer(Sint32 layer, float x, float y);
    Sint32 update(Sint32 layer = -1);
    Sint32 render(Sint32 layer = -1, Sint32 idRes = 0);

    Sint32 importLayerFromFile(Sint32 layer, const string& sFile, Sint32 tileSizeX, Sint32 tileSizeY, float accuracy = 100.0, Sint32 colorkey = -1);
    Sint32 importLayerFromImage(Sint32 layer, Sint32 idImg, Sint32 tileSizeX, Sint32 tileSizeY, float accuracy = 100.0, Sint32 colorkey = -1);
    Sint32 exportLayerToFile(Sint32 layer, const string& sFile);
    Sint32 exportLayerToImage(Sint32 layer, Sint32 idImg);
    Sint32 saveLevel(const string& sLevelFile, const string& sLevelName = "", const string& sCDCFile = "");
    Sint32 exportTMX(const string& sTMXFile, const string& sLevelName = "");

private:
    TileEngine(Uint32, const string&); // Heap-based and stack-based creation are forbidden.
    ~TileEngine(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};

//! TileEngine Manager class
class HandleDLL TileEngineMgr : public ManagerBase
{
    friend class Main;
public:
    Sint32 info(Sint32 iMode = 0);
    Sint32 create(const string& sName, Uint32 iNumLayers = 3);
    Sint32 close(Sint32 idTE);
    Sint32 getNum();
    Sint32 setName(Sint32 idTE, const string& sName);
    TileEngine* get(Sint32 idTE = 0);
    
    Sint32 loadLevel(const string& sLevelFile, const string& sLevelName = "", const string& sCDCFile = "");
    Sint32 importTMX(const string& sTMXFile);

private:
    TileEngineMgr(); // Heap-based and stack-based creation are forbidden.
    ~TileEngineMgr(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/

/** @addtogroup NetworkTCPM
@{*/
//! Network messages
enum eNetMsg
{
    NETMSG_NOTHING = 0, //!< No message received.        
    NETMSG_DATA = 64, //!< User data received, pData has it!
    NETMSG_DATA_ACCEPTED = 66, //!< Data sent by us was accepted, pData has it.
    NETMSG_DATA_DENIED = 67, //!< Data sent by us was denied, pData has the it.
    NETMSG_NEWCLIENT = 132, //!< New client connected, pData has the client name.
    NETMSG_QUITCLIENT = 133, //!< A client has exited, pData has the client name.
    NETMSG_CLOSE = 134, //!< Server has closed us.
    NETMSG_INFO = 135, //!< After request a fresh client info, it has arrived and ready to be check with getClientsInfo().
    NETMSG_PING = 136, //!< Periodic ping that the server send us.
    NETMSG_ERROR = 192 //!< Error, it usually means the client lost the connection with the server.
};

/** Used with Features() method. Disables the log into the server for new clients*/
#define NETFEATURE_DISABLE_LOGIN  1

//! NetTCP class
class HandleDLL NetTCP : public Base
{
    friend class Main;
public:
    //! Client information struct
    struct ClientInfo : public CMemAllocator
    {
        char szName[16]; //!< The name of the client. Updated automatically.
        Uint32 IP; //!< The IP v4 address of the client. Updated automatically.
        Sint32 iLatency; //!< The latency of the last second. Updated only when a client is connected or disconnected or you call to QueryClientsInfo().
        Sint32 iLatencyAvg; //!< The latency average of last 10 seconds Updated only when a client is connected or disconnected or you call to QueryClientsInfo().
    };

    Sint32 info(Sint32 iMode = 0);
    Sint32 init(Uint8 iLM);
    Sint32 close();
    Sint32 isServer();
    Sint32 isClient();
    Uint32 getIP();
    Sint32 getClientName(string& sClientName);
    Sint32 features(Sint32 iFeatures);

    // Server methods
    Sint32 createServer(Uint16 iPort, Sint32 iPasswd, Sint32 iDedicated);
    Sint32 setCoreServerCallback(Sint32 (*myCoreServer)(void* pData, Sint32 iSize, void* pObj), void* pObj = nullptr);

    // Client methods
    Sint32 connectTo(const string& sHost, Uint16 iPort, const string& sClient, Uint32 iPasswd);
    Sint32 sendData(void* pData, Sint32 iSize, Uint8 bIsQuery = 1);
    eNetMsg receiveData(void** pData, Uint32* iSize);
    Sint32 freeData(void*& buf);
    Sint32 queryKillServer();
    Sint32 queryKillClient();
    Sint32 queryClientsInfo();
    Sint32 getClientsInfo(ClientInfo** cinfo);
    Sint32 setTimeOut(Sint32 iMs);
    Sint32 setSimDelay(Sint32 iMs, Sint32 iMode = 0);

private:
    NetTCP(); // Heap-based and stack-based creation are forbidden.
    ~NetTCP(); // Inheritance is also forbidden.
    class PIMPL;
    PIMPL* mPData;
};
/** @}*/
}

#endif
